import click
import sys

import typing as tp

from robonomicsinterface import Account, constants, Datalog, Launch, SubEvent, Subscriber


def callback(data: tp.Tuple[tp.Union[str, int]]) -> None:
    """
    callback executed when subscription event triggered. Simply outputs incoming info to console

    :param data: data to be output

    """
    click.echo(data)


@click.group()
def cli() -> None:
    pass


@cli.group(help="Send various extrinsics (launch commands or record datalogs)")
def write() -> None:
    pass


@cli.group(help="Subscribe to datalog/launch events in the chain")
def read() -> None:
    pass


@write.command()
@click.option(
    "--input_string",
    type=click.File("r"),
    default=sys.stdin,
    hidden=True,
    help="Hidden parameter to perform stdin reading of a passed via pipeline sting",
)
@click.option(
    "--remote_ws",
    type=str,
    default=constants.REMOTE_WS,
    help="Node websocket address used to connect to any node. E.g. local is ws://127.0.0.1:9944. Default is "
    "wss://kusama.rpc.robonomics.network",
)
@click.option("-s", type=str, required=True, help="Account seed in mnemonic/raw form.")
def datalog(input_string: sys.stdin, remote_ws: str, s: str) -> None:
    """
    Save string into account's datalog using pipeline:  <echo "blah" | robonomics_interface io write datalog (params)>
    If nothing passed, waits for a string in a new line.
    """
    account: Account = Account(remote_ws=remote_ws, seed=s)
    datalog_: Datalog = Datalog(account)
    transaction_hash: str = datalog_.record(input_string.readline()[:-1])
    click.echo(transaction_hash)


@write.command()
@click.option(
    "--command",
    type=click.File("r"),
    default=sys.stdin,
    hidden=True,
    help="Hidden parameter to perform stdin reading of a passed via pipeline command",
)
@click.option(
    "--remote_ws",
    type=str,
    default=constants.REMOTE_WS,
    help="Node websocket address used to connect to any node. E.g. local is ws://127.0.0.1:9944. Default is "
    "wss://kusama.rpc.robonomics.network",
)
@click.option("-s", type=str, required=True, help="Account seed in mnemonic/raw form.")
@click.option("-r", type=str, required=True, help="Target account ss58_address.")
def launch(command: sys.stdin, remote_ws: str, s: str, r: str) -> None:
    """
    Send launch command accompanied by parameter in IPFS Qm... form or just 32 bytes data using pipeline:
    <echo "Qmc5gCcjYypU7y28oCALwfSvxCBskLuPKWpK4qpterKC7z" | robonomics_interface io write launch (params)>
    If nothing passed, waits for a string in a new line.
    """
    account: Account = Account(remote_ws=remote_ws, seed=s)
    launch_: Launch = Launch(account)
    parameter: str = command.readline()[:-1]
    transaction_hash: str = launch_.launch(r, parameter)
    click.echo((transaction_hash, f"{account.get_address()} -> {r}: {parameter}"))


@read.command()
@click.option(
    "--remote_ws",
    type=str,
    default=constants.REMOTE_WS,
    help="Node websocket address used to connect to any node. E.g. local is ws://127.0.0.1:9944. Default is "
    "wss://kusama.rpc.robonomics.network",
)
@click.option("-r", type=str, help="Target account ss58_address.")
def datalog(remote_ws: str, r: str) -> None:
    """
    Listen to datalogs in the chain whether address-specified or all of them
    """
    account: Account = Account(remote_ws=remote_ws)
    subscriber: Subscriber = Subscriber(account, SubEvent.NewRecord, subscription_handler=callback, addr=r)
    pass


@read.command()
@click.option(
    "--remote_ws",
    type=str,
    default=constants.REMOTE_WS,
    help="Node websocket address used to connect to any node. E.g. local is ws://127.0.0.1:9944. Default is "
    "wss://kusama.rpc.robonomics.network",
)
@click.option("-r", type=str, help="Target account ss58_address.")
def launch(remote_ws: str, r: str) -> None:
    """
    Listen to datalogs in the chain whether address-specified or all of them
    """
    account: Account = Account(remote_ws=remote_ws)
    subscriber: Subscriber = Subscriber(account, SubEvent.NewLaunch, subscription_handler=callback, addr=r)
    pass


if __name__ == "__main__":
    cli()



import hashlib
import logging
import typing as tp

import substrateinterface as substrate

from base58 import b58decode, b58encode
from scalecodec.base import RuntimeConfiguration, ScaleBytes, ScaleType

logger = logging.getLogger(__name__)


def create_keypair(seed: str) -> substrate.Keypair:
    """
    Create a keypair for further use.

    :param seed: Account seed (mnemonic or raw) as a key to sign transactions.

    :return: A Keypair instance used by substrate to sign transactions.

    """

    if seed.startswith("0x"):
        return substrate.Keypair.create_from_seed(seed_hex=hex(int(seed, 16)), ss58_format=32)
    else:
        return substrate.Keypair.create_from_mnemonic(seed, ss58_format=32)


def dt_encode_topic(topic: str) -> str:
    """
    Encode any string to be accepted by Digital Twin setSource. Use byte encoding and sha256-hashing.

    :param topic: Topic name to be encoded.

    :return: Hashed-encoded topic name

    """

    return f"0x{hashlib.sha256(topic.encode('utf-8')).hexdigest()}"


def ipfs_32_bytes_to_qm_hash(string_32_bytes: str) -> str:
    """
    Transform 32 bytes sting (without 2 heading bytes) to an IPFS base58 Qm... hash.

    :param string_32_bytes: 32 bytes sting (without 2 heading bytes).

    :return: IPFS base58 Qm... hash.

    """

    if string_32_bytes.startswith("0x"):
        string_32_bytes = string_32_bytes[2:]
    return b58encode(b"\x12 " + bytes.fromhex(string_32_bytes)).decode("utf-8")


def ipfs_qm_hash_to_32_bytes(ipfs_qm: str) -> str:
    """
    Transform IPFS base58 Qm... hash to a 32 bytes sting (without 2 heading '0x' bytes).

    :param ipfs_qm: IPFS base58 Qm... hash.

    :return: 32 bytes sting (without 2 heading bytes).

    """

    return f"0x{b58decode(ipfs_qm).hex()[4:]}"


def str_to_scalebytes(data: tp.Union[int, str], type_str: str) -> ScaleBytes:
    """
    Encode string to a desired ScaleBytes data.

    :param data: String to encode.
    :param type_str: Type (``U32``, ``Compact<Balance>``, etc.).

    :return: ScaleBytes object

    """

    scale_obj: ScaleType = RuntimeConfiguration().create_scale_object(type_str)
    return scale_obj.encode(data)



import substrateinterface as substrate

from functools import wraps
from websocket._exceptions import WebSocketConnectionClosedException

def check_socket_opened(func):
    """
    Open and substrate node connection each time needed.

    :param func: wrapped function.

    :return: wrapped function after augmentations.

    """

    @wraps(func)
    def wrapper(ri_instance, *args, **kwargs):
        """
        Wrap decorated function with interface opening if it was closed.

        :param ri_instance: RobonomicsInterface instance in a decorated function.
        :param args: Wrapped function args.
        :param kwargs: Wrapped function kwargs.

        """

        if not ri_instance.interface:
            open_interface(ri_instance)

        try:
            res = func(ri_instance, *args, **kwargs)
        except (BrokenPipeError, WebSocketConnectionClosedException):
            open_interface(ri_instance)
            res = func(ri_instance, *args, **kwargs)

        return res

    return wrapper


def open_interface(ri_instance):

    ri_instance.interface = substrate.SubstrateInterface(
        url=ri_instance.remote_ws,
        ss58_format=32,
        type_registry_preset="substrate-node-template",
        type_registry=ri_instance.type_registry,
    )



import typing as tp

AccountTyping = tp.Dict[str, tp.Union[int, tp.Dict[str, int]]]
AuctionTyping = tp.Dict[str, tp.Union[str, int, tp.Dict[str, tp.Dict[str, tp.Dict[str, int]]]]]
DatalogTyping = tp.Tuple[int, tp.Union[int, str]]
DigitalTwinTyping = tp.List[tp.Tuple[str, str]]
LedgerTyping = tp.Dict[str, tp.Union[int, tp.Dict[str, tp.Dict[str, tp.Dict[str, int]]]]]
LiabilityTyping = tp.Dict[str, tp.Union[tp.Dict[str, tp.Union[str, int]], str]]
ListenersResponse = tp.Dict[str, tp.Union[str, tp.List[str], int]]
QueryParams = tp.Optional[tp.Union[tp.List[tp.Union[str, int]], str, int]]
ReportTyping = tp.Dict[str, tp.Union[int, str, tp.Dict[str, str]]]
RWSParamsTyping = tp.Dict[str, tp.Union[str, tp.Dict[str, tp.Union[str, int, dict, list]]]]
TypeRegistryTyping = tp.Dict[str, tp.Dict[str, tp.Union[str, tp.Any]]]



import typing as tp

from logging import getLogger

from .base import BaseClass
from ..types import AccountTyping

logger = getLogger(__name__)


class CommonFunctions(BaseClass):
    """
    Class for common functions such as getting account information or transferring tokens
    """

    def get_account_info(self, addr: tp.Optional[str] = None, block_hash: tp.Optional[str] = None) -> AccountTyping:
        """
        Get account information.

        :param addr: Explored account ss58 address. Account dataclass address if None.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Account information dictionary.

        """

        account_address: str = addr or self.account.get_address()

        logger.info(f"Getting account {account_address} data")

        return self._service_functions.chainstate_query("System", "Account", account_address, block_hash=block_hash)

    def get_account_nonce(self, addr: tp.Optional[str] = None) -> int:
        """
        Get current account nonce.

        :param addr: Account ss58 address. Self address via private key is obtained if not passed.

        :return Account nonce. Due to the feature of substrate-interface lib, to create an extrinsic with incremented
            nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        """

        account_address: str = addr or self.account.get_address()

        logger.info(f"Fetching nonce of account {account_address}")

        return self._service_functions.rpc_request(
            "system_accountNextIndex", [account_address], result_handler=None
        ).get("result", 0)

    def transfer_tokens(self, target_address: str, tokens: int, nonce: tp.Optional[int] = None) -> str:
        """
        Send tokens to target address.

        :param target_address: Account that will receive tokens.
        :param tokens: Number of tokens to be sent, in Wei, so if you want to send 1 XRT, you should send
            "1 000 000 000" units.
        :param nonce: Account nonce. Due to the feature of substrate-interface lib,
            to create an extrinsic with incremented nonce, pass account's current nonce. See
            https://github.com/polkascan/py-substrate-interface/blob/85a52b1c8f22e81277907f82d807210747c6c583/substrateinterface/base.py#L1535
            for example.

        :return: Hash of the transfer transaction.

        """

        logger.info(f"Sending tokens to {target_address}")

        return self._service_functions.extrinsic(
            "Balances",
            "transfer",
            {"dest": {"Id": target_address}, "value": tokens},
            nonce,
        )



import typing as tp

from logging import getLogger

from .base import BaseClass
from ..types import AuctionTyping, LedgerTyping

logger = getLogger(__name__)


class RWS(BaseClass):
    """
    Class for interacting with Robonomics Web Services subscriptions
    """

    def get_auction(self, index: int, block_hash: tp.Optional[str] = None) -> tp.Optional[AuctionTyping]:
        """
        Get information about subscription auction.

        :param index: Auction index.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Auction info.

        """

        logger.info(f"Fetching auction {index} information")
        return self._service_functions.chainstate_query("RWS", "Auction", index, block_hash=block_hash)

    def get_auction_next(self, block_hash: tp.Optional[str] = None) -> int:
        """
        Get index of the next auction to be unlocked.

        :param block_hash: Retrieves data as of passed block hash.

        :return: Auction index.

        """

        logger.info("Fetching index of the next auction to be unlocked")
        return self._service_functions.chainstate_query("RWS", "AuctionNext", block_hash=block_hash)

    def get_auction_queue(self, block_hash: tp.Optional[str] = None) -> tp.List[tp.Optional[int]]:
        """
        Get an auction queue of Robonomics Web Services subscriptions.

        :param block_hash: Retrieves data as of passed block hash.

        :return: Auction queue of Robonomics Web Services subscriptions.

        """

        logger.info("Fetching auctions queue list")
        return self._service_functions.chainstate_query("RWS", "AuctionQueue", block_hash=block_hash)

    def get_devices(
        self, addr: tp.Optional[str] = None, block_hash: tp.Optional[str] = None
    ) -> tp.List[tp.Optional[str]]:
        """
        Fetch list of RWS added devices.

        :param addr: Subscription owner. If ``None`` - account address.
        :param block_hash: Retrieves data as of passed block hash.

        :return: List of added devices. Empty if none.

        """

        address: str = addr or self.account.get_address()

        logger.info(f"Fetching list of RWS devices set by owner {address}")

        return self._service_functions.chainstate_query("RWS", "Devices", address, block_hash=block_hash)

    def get_ledger(
        self, addr: tp.Optional[str] = None, block_hash: tp.Optional[str] = None
    ) -> tp.Optional[LedgerTyping]:
        """
        Subscription information.

        :param addr: Subscription owner. If ``None`` - account address.
        :param block_hash: Retrieves data as of passed block hash.

        :return: Subscription information. Empty if none.

        """

        address: str = addr or self.account.get_address()

        logger.info(f"Fetching subscription information by owner {address}")

        return self._service_functions.chainstate_query("RWS", "Ledger", address, block_hash=block_hash)

    def bid(self, index: int, amount: int) -> str:
        """
        Bid to win a subscription!

        :param index: Auction index.
        :param amount: Your bid in Weiners.

        :return: Transaction hash.

        """

        logger.info(f"Bidding on auction {index} with {amount} Weiners (appx. {round(amount / 10 ** 9, 2)} XRT)")
        return self._service_functions.extrinsic("RWS", "bid", {"index": index, "amount": amount})

    def set_devices(self, devices: tp.List[str]) -> str:
        """
        Set devices which are authorized to use RWS subscriptions held by the extrinsic author.

        :param devices: Devices authorized to use RWS subscriptions. Include in list.

        :return: Transaction hash.

        """

        logger.info(f"Allowing {devices} to use {self.account.get_address()} subscription")
        return self._service_functions.extrinsic("RWS", "set_devices", {"devices": devices})



import typing as tp

from dataclasses import dataclass
from logging import getLogger
from substrateinterface import Keypair

from ..constants import REMOTE_WS, TYPE_REGISTRY
from ..exceptions import NoPrivateKeyException
from ..types import TypeRegistryTyping
from ..utils import create_keypair

logger = getLogger(__name__)


@dataclass
class Account:
    """
    Dataclass to hold account info and node connection parameters

    """

    def __init__(
        self,
        seed: tp.Optional[str] = None,
        remote_ws: tp.Optional[str] = None,
        type_registry: tp.Optional[TypeRegistryTyping] = None,
    ) -> None:
        """
        Save node connection parameters and create a keypair to sign transactions and define address if seed was passed
            as a parameter.

        :param seed: Account seed (mnemonic or raw) as a key to sign transactions.
        :param remote_ws: Node url. Default node address is "wss://kusama.rpc.robonomics.network". Another address may
            be specified (e.g. "ws://127.0.0.1:9944" for local node).
        :param type_registry: Types used in the chain. Defaults are the most frequently used in Robonomics.

        """
        self.remote_ws: tp.Optional[str] = remote_ws or REMOTE_WS
        self.type_registry: tp.Optional[TypeRegistryTyping] = type_registry or TYPE_REGISTRY
        if seed:
            self.keypair: tp.Optional[Keypair] = create_keypair(seed)
            self._address: str = self.keypair.ss58_address
        else:
            self.keypair: tp.Optional[Keypair] = None

    def get_address(self) -> str:
        """
        Determine account address if seed was passed when creating an instance

        :return: Account ss58 address

        """
        if not self.keypair:
            raise NoPrivateKeyException("No private key was provided, unable to determine account address")
        return str(self.keypair.ss58_address)



import typing as tp

from enum import Enum
from logging import getLogger
from websocket import WebSocketConnectionClosedException

from .account import Account
from .service_functions import ServiceFunctions

logger = getLogger(__name__)


class SubEvent(Enum):
    NewRecord = "NewRecord"
    NewLaunch = "NewLaunch"
    Transfer = "Transfer"
    TopicChanged = "TopicChanged"
    NewDevices = "NewDevices"


class Subscriber:
    """
    Class intended for use in cases when needed to subscribe on chainstate updates/events. **Blocks current thread**!
    """

    def __init__(
        self,
        account: Account,
        subscribed_event: SubEvent,
        subscription_handler: callable,
        addr: tp.Optional[tp.Union[tp.List[str], str]] = None,
    ) -> None:
        """
        Initiates an instance for further use and starts a subscription for a selected action.

        :param account: Account dataclass with ``seed``, ``remote_ws`` address and node ``type_registry``.
        :param subscribed_event: Event in substrate chain to be awaited. Choose from [``NewRecord``, ``NewLaunch``,
            ``Transfer``, ``TopicChanged``, ``NewDevices``]. This parameter should be a ``SubEvent`` class attribute.
            This also requires importing this class.
        :param subscription_handler: Callback function that processes the updates of the storage.
            THIS FUNCTION IS MEANT TO ACCEPT ONLY ONE ARGUMENT (THE NEW EVENT DESCRIPTION TUPLE).
        :param addr: ss58 type 32 address(-es) of an account(-s) which is(are) meant to be event target. If ``None``,
            will subscribe to all such events never-mind target address(-es).

        """

        self._custom_functions: ServiceFunctions = ServiceFunctions(account)
        self._event: SubEvent = subscribed_event
        self._callback: callable = subscription_handler
        self._target_address: tp.Optional[tp.Union[tp.List[str], str]] = addr

        self._subscribe_event()

    def _subscribe_event(self) -> None:
        """
        Subscribe to events targeted to a certain account (``launch``, ``transfer``). Call ``subscription_handler``
        when updated.

        """

        logger.info(f"Subscribing to event {self._event.value} for target addresses {self._target_address}")
        try:
            self._custom_functions.subscribe_block_headers(self._event_callback)
        except WebSocketConnectionClosedException:
            self._subscribe_event()

    def _event_callback(self, index_obj: tp.Any, update_nr: int, subscription_id: int) -> None:
        """
        Function, processing updates in event list storage. On update filters events to a desired account
        and passes the event description to the user-provided ``callback`` method.

        :param index_obj: Updated event list.
        :param update_nr: Update counter. Increments every new update added. Starts with ``0``.
        :param subscription_id: Subscription ID.

        """

        if update_nr == 0:
            return None

        chain_events: list = self._custom_functions.chainstate_query("System", "Events")
        for events in chain_events:
            if events["event_id"] == self._event.value:
                if self._target_address is None:
                    self._callback(events["event"]["attributes"])  # All events
                elif (
                    events["event"]["attributes"][
                        0 if self._event in [SubEvent.NewRecord, SubEvent.TopicChanged, SubEvent.NewDevices] else 1
                    ]
                    in self._target_address
                ):
                    self._callback(events["event"]["attributes"])  # address-targeted
